from abc import ABC, abstractmethod
from enum import Enum
from typing import Any, Iterable, Optional, Type, Union
from warnings import warn


class Solver(ABC):

    class Sense(Enum):
        MAX = 1
        MIN = 2

    @abstractmethod
    def addVariable(self, name: str, vartype: str, obj: Optional[float],
                    lb: Optional[float], ub: Optional[float]):
        """Adds a variable to the model

        Args:
            name (str): Name of the variable, used to reference the variable and for printing
            vartype (str): Type of the variable ('B' - Binary, 'C' - continuous, 'I' - Integer)
            obj (Optional[float]): Coefficient at the variable in the objective function, default is 0.
            lb (Optional[float]): Lower bound of the variable, defaults are dictated by the solver.
            ub (Optional[float]): Upper bound of the variable, defaults are dictated by the solver.
        Returns:
            vartype: An object representing the variable in the selected solver.
            It has to be used in the 'variables' arg in the addConstraint method
        """
        pass

    @abstractmethod
    def addConstraint(self, name: str, variables: Iterable[Any],
                      coeffs: list[float], rhs: float, sense: str):
        """Add a constraint to the model.

        In general, the constraints have to be in the form of:
        <linear combination of variables> < 'L'|'G'|'E' > <independent part of the expression>
        , where 'L'|'G'|'E' are 'less than or equal' | 'greater than or equal' | 'equal' respectively.

        For example:

        2*x1 + 3*x2 >= 8

        would be created by:

        model.addConstraint('example', [x1, x2], [2, 3], 8, 'G')

        Args:
            name (str): Name of the constraint, useful when printing the model.
            variables (list[object]): List of the variables present in the left-hand side of the constraint.
            coeffs (list[float]): Coefficients at the variables, must of the same length and in the same order as <variables>.
            rhs (float): The independent part of the expression
            sense (str): Relation between lhs and rhs, can be one of 'L'|'G'|'E'
        """
        pass

    @abstractmethod
    def solve(self):
        """Solves the model
        """
        pass

    @abstractmethod
    def getValues(self):
        """Get values of all the model variables, in the same order they were added to the model.
        """
        pass


class CPLEX(Solver):

    def __init__(self,
                 sense: Solver.Sense = Solver.Sense.MAX,
                 log: bool = False) -> None:
        super().__init__()
        try:
            import cplex
        except ImportError as err:
            raise ImportError(
                "Failed to import CPLEX, please make sure that you have CPLEX installed"
            )

        self.col_names = []
        self.obj = []
        self.lb = []
        self.ub = []
        self.ctypes = []

        self.rhs = []
        self.row_names = []
        self.senses = []
        self.rows = []

        self.cplex = cplex
        self.model = cplex.Cplex()
        if not log:
            self.model.set_log_stream(None)
            self.model.set_results_stream(None)

        if sense == Solver.Sense.MAX:
            self.model.objective.set_sense(self.model.objective.sense.maximize)
        elif sense == Solver.Sense.MIN:
            self.model.objective.set_sense(self.model.objective.sense.minimize)

    def addVariable(self,
                    name: str,
                    vartype: str,
                    obj: float = 0.0,
                    lb: Optional[float] = None,
                    ub: Optional[float] = None):
        self.col_names.append(name)
        self.obj.append(obj)
        if lb is not None:
            self.lb.append(lb)
        if ub is not None:
            self.ub.append(ub)
        self.ctypes.append(vartype)
        return name

    def addConstraint(self, name: str, variables: Iterable[Any],
                      coeffs: list[float], rhs: float, sense: str):
        self.row_names.append(name)
        self.rows.append([variables, coeffs])
        self.rhs.append(rhs)
        self.senses.append(sense)

    def solve(self):
        if len(self.lb) != len(self.ctypes) or len(self.ub) != len(
                self.ctypes):
            if len(self.lb) > 0 or len(self.ub) > 0:
                warn(
                    "Warning: You've given bounds for some variables but not for all - please make sure you either set bounds for every variable or for none.\nUsing default bounds..."
                )
            self.model.variables.add(obj=self.obj,
                                     types=''.join(self.ctypes),
                                     names=self.col_names)
        else:
            self.model.variables.add(obj=self.obj,
                                     lb=self.lb,
                                     ub=self.ub,
                                     types=''.join(self.ctypes),
                                     names=self.col_names)
        self.model.linear_constraints.add(lin_expr=self.rows,
                                          senses=''.join(self.senses),
                                          rhs=self.rhs,
                                          names=self.row_names)
        self.model.solve()

    def getValues(self):
        return self.model.solution.get_values()


class Gurobi(Solver):

    def __init__(self,
                 sense: Solver.Sense = Solver.Sense.MAX,
                 log: bool = False) -> None:
        super().__init__()
        try:
            import gurobipy as gp
            self.gp = gp
            from gurobipy import GRB
        except ImportError as _err:
            raise ImportError(
                "Failed to import Gurobi, please make sure that you have Gurobi installed"
            )

        # TODO
        # forcefully disabling Gurobi logs for now, can be replaced by config file later
        env = gp.Env(empty=True)
        if not log:
            env.setParam('OutputFlag', 0)
        env.start()

        self.model = gp.Model(env=env)
        self.var_type_map = {
            'C': GRB.CONTINUOUS,
            'B': GRB.BINARY,
            'I': GRB.INTEGER
        }

        self.objective = []
        if sense == Solver.Sense.MAX:
            self.sense = GRB.MAXIMIZE
        elif sense == Solver.Sense.MIN:
            self.sense = GRB.MINIMIZE

    def addVariable(self,
                    name: str,
                    var_type: str,
                    obj: Optional[float] = None,
                    lb: Optional[float] = None,
                    ub: Optional[float] = None):
        var = self.model.addVar(vtype=self.var_type_map[var_type], name=name)
        if lb is not None:
            self.model.addConstr(var >= lb, name + '_lb')
        if ub is not None:
            self.model.addConstr(var <= ub, name + '_ub')
        if obj != None:
            self.objective.append(obj * var)
        return var

    def addConstraint(self, name: str, variables: Iterable[Any],
                      coeffs: list[float], rhs: float, sense: str):
        expr = sum([v * c for v, c in zip(variables, coeffs)])
        if sense == 'L':
            self.model.addConstr(expr <= rhs, name)
        elif sense == 'G':
            self.model.addConstr(expr >= rhs, name)
        elif sense == 'E':
            self.model.addConstr(expr == rhs, name)

    def _setObjective(self):
        self.model.setObjective(sum(self.objective), self.sense)

    def solve(self):
        self._setObjective()
        self.model.optimize()

    def getValues(self):
        return [v.X for v in self.model.getVars()]

    def _debug(self, filename='model_dump.lp'):
        self._setObjective()
        try:
            self.model.computeIIS()
            print("Computed Irreducible Inconsistent Subsystem:")
            for i, c in zip(self.model.IISConstr, self.model.getConstrs()):
                if i == 1:
                    print(c)
        except self.gp.GurobiError:
            print("The model is feasible!")
        self.model.write(filename)


solver_t = Union[Type[CPLEX], Type[Gurobi]]
model_t = Union[CPLEX, Gurobi]



import collections

from votepy.algorithms.p_algorithm import PAlgorithm
from votepy.ordinal_election import OrdinalElection

from votepy.meta.structure import rule, impl
from votepy.solve import solve

from votepy.algorithms.base_algorithm import BaseAlgorithm
from votepy.algorithms.brute_force import BruteForce
from votepy.algorithms.greedy import Greedy
from votepy.algorithms.basin_hopping import BasinHopping
from votepy.algorithms.ilp import ILP
from votepy.generic_ilp import CPLEX, Gurobi, model_t, solver_t

from typing import Type, Union, Iterable


@rule()
def chamberlin_courant(voting: Union[list[list[int]], OrdinalElection], size_of_committee: int, algorithm: BaseAlgorithm) -> \
        list[int]:
    """# Summary
    Chamberlin-Courant rule
    ## Args:
        `voting` (`list[list[int]]` | `OrdinalElection`): Voting for which the function calculates the committee
        `size_of_committee` (`int`): Size of the committee

    ## Returns:
        `list[int]`: List of chosen candidates

    ## Examples
    >>> election = [
    ...     [0, 1, 2, 3, 4],
    ...     [4, 0, 1, 3, 2],
    ...     [3, 0, 1, 2, 4],
    ...     [2, 1, 3, 4, 0],
    ...     [2, 1, 4, 0, 3],
    ...     [1, 2, 3, 4, 0]
    ... ]
    >>> chamberlin_courant(election, 2, algorithm='brute_force')
    [0, 2]
    >>> chamberlin_courant(election, 2, algorithm='greedy')
    [1, 0]
    """
    return solve(chamberlin_courant, voting, size_of_committee, algorithm=algorithm)


def scoring_function(committee: Iterable[int], voting: OrdinalElection) -> float:
    """# Summary
    Scoring function for Chamberlin-Courant rule. Calculates score for the given committee, under CC rule and given voting

    ## Args:
        `committee` (Iterable[int]): Chosen committee
        `voting` (OrdinalElection): Voting

    ## Returns:
        -> float: Score of the committee

    ## Examples
    """
    committee = set(committee)
    score = 0
    m = voting.ballot_size
    for voter in voting:
        score += max(m - voter.pos(candidate) for candidate in committee)
    return score


@impl(chamberlin_courant, BruteForce)
def chamberlin_courant_brute_force(voting, size_of_committee, algorithm: BruteForce = BruteForce()) -> list[int]:
    algorithm.prepare(scoring_function)
    return algorithm.solve(voting, size_of_committee)


@impl(chamberlin_courant, Greedy)
def chamberlin_courant_greedy(voting, size_of_committee, algorithm: Greedy = Greedy()) -> list[int]:
    def greedy_scoring_function(committee: Iterable[int], voting: OrdinalElection, candidate: int):
        committee = set(committee) | {candidate}
        return scoring_function(committee, voting)

    algorithm.prepare(greedy_scoring_function)
    return algorithm.solve(voting, size_of_committee)


# Based on https://arxiv.org/abs/1901.09217
@impl(chamberlin_courant, PAlgorithm)
def chamberlin_courant_p_algorithm(voting: Union[list[list[int]], OrdinalElection], size_of_committee: int,
                                   algorithm: PAlgorithm = PAlgorithm()) -> list[int]:
    def scoring_function(voting: OrdinalElection, size_of_committee: int, threshold: int) -> list[int]:
        scores = collections.defaultdict(int)
        winners = []
        vs = [v[:threshold] for v in voting]
        for vote in vs:
            for candidate in vote:
                scores[candidate] += 1
        for _ in range(size_of_committee):
            best_candidate = max((val, key) for key, val in scores.items())[1]

            for idx, cs in enumerate(vs):
                if best_candidate in cs:
                    for c in cs:
                        scores[c] -= 1
                    vs[idx] = []
            del scores[best_candidate]
            winners.append(best_candidate)
        return winners

    algorithm.prepare(scoring_function)
    return algorithm.solve(voting, size_of_committee)


@impl(chamberlin_courant, BasinHopping)
def chamberlin_courant_basinhopping(voting: Union[list[list[int]], OrdinalElection], size_of_committee: int,
                                    algorithm: BasinHopping = BasinHopping()):
    algorithm.prepare(scoring_function)
    return algorithm.solve(voting, size_of_committee)


@impl(chamberlin_courant, ILP)
def chamberlin_courant_ilp(voting: Union[list[list[int]], OrdinalElection], size_of_committee: int, algorithm: ILP = ILP(Gurobi)) -> list[int]:
    """Implementation of the chamberlin-courant rule, using ILP formulation by:
    Peters, Dominik & Lackner, Martin. (2020).
    Preferences Single-Peaked on a Circle.
    Journal of Artificial Intelligence Research. 68. 463-502. 10.1613/jair.1.11732.

    Args:
        voting (Union[list[list[int]], OrdinalElection]): Voting for which the function calculates the committee
        size_of_committee (`int`): Size of the committee
    Returns:
        list[int]: List of chosen candidates
    """
    def define_model(voting: OrdinalElection, size_of_committee: int, solver: solver_t) -> model_t:
        model = solver()

        y = [model.addVariable(f"y_{c}", 'B', 0)
             for c in range(voting.ballot_size)]

        x = [[model.addVariable(f"x_{i},{r}", 'B', 1)
              for r in range(voting.ballot_size)]
             for i in range(voting.number_of_voters)]

        model.addConstraint(
            'sum(y)', y, [1]*len(y), size_of_committee, 'E')

        for i in range(voting.number_of_voters):
            for r in range(voting.ballot_size):
                top_r_candidates = [y[c] for c in voting[i][:r]]
                model.addConstraint(
                    f'x_{i}_{r} <= sum(top_{r}_candidates)',
                    [x[i][r]] + top_r_candidates,
                    [1.0] + [-1.0]*len(voting[i][:r]),
                    0,
                    'L'
                )
        return model

    algorithm.prepare(define_model)
    return algorithm.solve(voting, size_of_committee)


# For now leaving without annotation - if it's easy to add different
# implementations of the same rule and algorithm, then I think we should do
# that. To use this function, you need to call it explicitly.
def chamberlin_courant_ilp_custom(voting: Union[list[list[int]], OrdinalElection], size_of_committee: int, algorithm: ILP = ILP(Gurobi)) -> list[int]:
    """Custom implementation of the chamberlin courant rule.

    Args:
        voting (Union[list[list[int]], OrdinalElection]): Voting for which the function calculates the committee
        size_of_committee (`int`): Size of the committee
    Returns:
        list[int]: List of chosen candidates
    """

    def define_model(voting: OrdinalElection, size_of_committee: int, solver: solver_t) -> model_t:
        model = solver()
        x = [model.addVariable(f"x_{i}", 'B') for i in range(voting.ballot_size)]

        y = []
        for i, vote in enumerate(voting):
            y.append([None]*voting.ballot_size)
            for j, candidate in enumerate(vote):
                y[-1][candidate] = (model.addVariable(
                    f"y_{i}_{j}", "B", voting.ballot_size - j))

        model.addConstraint("sum(x) == k", x, [1]*len(x), size_of_committee, 'E')

        for i in range(voting.number_of_voters):
            model.addConstraint(f"sum(y[{i}]) == 1", y[i], [1]*len(y[i]), 1, 'E')
            for j in range(voting.ballot_size):
                model.addConstraint(f"x_{j} >= y_{i}_{j}", [
                                    x[j], y[i][j]], [1, -1], 0, 'G')

        return model

    algorithm.prepare(define_model)
    return algorithm.solve(voting, size_of_committee)


if __name__ == '__main__':
    election = OrdinalElection([
        [0, 1, 2, 3, 4],
        [4, 0, 1, 3, 2],
        [3, 0, 1, 2, 4],
        [2, 1, 3, 4, 0],
        [2, 1, 4, 0, 3],
        [1, 2, 3, 4, 0]
    ], {
        0: 'a',
        1: 'b',
        2: 'c',
        3: 'd',
        4: 'e',
    })
    print(election)
    print(
        chamberlin_courant(
            election,
            2,
            algorithm='brute_force'
        )
    )
    print(
        chamberlin_courant(
            election,
            2,
            algorithm=Greedy()
        )
    )
    print(
        chamberlin_courant(
            election,
            2,
            algorithm=PAlgorithm()
        )
    )
    print(
        chamberlin_courant(
            election,
            2,
            algorithm=ILP(Gurobi)
        )
    )
    print(
        chamberlin_courant(
            election,
            2,
            algorithm=ILP(CPLEX)
        )
    )

    print(
        chamberlin_courant_ilp_custom(
            election,
            2,
            algorithm=ILP(Gurobi)
        )
    )
    print(
        chamberlin_courant_ilp_custom(
            election,
            2,
            algorithm=ILP(CPLEX)

        )
    )



from votepy.algorithms.base_algorithm import BaseAlgorithm
from votepy.algorithms.brute_force import BruteForce
from votepy.algorithms.greedy import Greedy
from votepy.algorithms.p_algorithm import PAlgorithm
from votepy.algorithms.basin_hopping import BasinHopping
from votepy.algorithms.ilp import ILP



from typing import Callable, Iterable, Union
from votepy.ordinal_election import OrdinalElection
from votepy.algorithms.base_algorithm import BaseAlgorithm
from votepy.meta.structure import get_algorithm, get_default_algorithm, get_implementation, has_default_implementation


def solve(rule: Union[Callable, str], voting: Union[list[list[int]], OrdinalElection], size_of_committee: int, *rule_args, algorithm: BaseAlgorithm = None, **rule_kwargs) -> list[int]:
    """# Summary
    Calculates committee of the given size, based on the voting rule, using algorithm.

    ## Args:
        `rule` (Callable | str): Voting rule function or its identification name
        `voting` (Union[list[list[int]], OrdinalElection]): Voting for which the function calculates the committee
        `size_of_committee` (`int`): Size of the committee
        `algorithm` (BaseAlgorithm): Voting solving algorithm or its identification name.
        `*rule_args`: Additional voting functions' arguments
        `*rule_kwargs`: Additional voting functions' positional arguments

    ## Returns:
        -> list[int]: List of chosen candidates

    ## Examples
    >>> solve('k_borda', [
    ...     [0, 1, 2, 3],
    ...     [3, 2, 1, 0],
    ...     [2, 1, 3, 0]
    ... ], 2)
    [2, 1]
    """
    if algorithm is None:
        if not has_default_implementation(rule):
            raise ValueError(
                f"{rule} has no default implementation. If rule does not need any algorithm, its rule function should be additionally decorated with `@impl(rule='name', algorithm=None)`. See `@impl` docs.")
        algorithm = get_default_algorithm(rule)

    implementation = get_implementation(rule, algorithm)
    algorithm = get_algorithm(algorithm)

    if algorithm is None:
        result = implementation(voting, size_of_committee, *rule_args, **rule_kwargs)
    else:
        result = implementation(voting, size_of_committee, *rule_args, algorithm=algorithm, **rule_kwargs)

    if isinstance(result, Iterable):
        return list(result)
    else:
        return result


if __name__ == '__main__':
    from votepy.rules.k_borda_rule import k_borda
    solve(k_borda, [
        [0, 1, 2, 3],
        [3, 2, 1, 0],
        [2, 1, 3, 0]
    ], 2, 3)



from votepy.ordinal_election import OrdinalElection

__meta_tokens = {
    'FILE NAME',
    'TITLE',
    'DESCRIPTION',
    'DATA TYPE',
    'MODIFICATION TYPE',
    'RELATES TO',
    'RELATED FILES',
    'PUBLICATION DATE',
    'MODIFICATION DATE',
    'NUMBER ALTERNATIVES',

    'NUMBER VOTERS',
    'NUMBER UNIQUE ORDERS',
}


def load(source: str) -> OrdinalElection:
    """# Summary
    Loads election from file in the preflib format.

    ## Args:
        `source` (str): Path to the file with voting

    ## Returns:
        ->OrdinalElection: Election

    ## Examples
    >>> load('tests/fixtures/preflib/voting_no_dups.txt')
    [[0, 1, 2, 3], [3, 2, 1, 0], [2, 1, 3, 0]]
    """
    with open(source, 'r') as f:
        lines = f.readlines()
        mapping = {}

        election = []

        for line in lines:
            if line.isspace():
                continue

            if line.startswith('# '):
                token, value = line[2:].split(': ')
                if token.startswith('ALTERNATIVE NAME'):
                    _, _, alternative_id = token.split()
                    mapping[int(alternative_id) - 1] = value.strip()
            else:
                n_voters, ballot = line.split(': ')
                n_voters = int(n_voters)
                ballot = ballot.replace('{', '').replace('}', '').strip().split(',')
                ballot = list(map(lambda x: int(x) - 1, ballot))
                election.extend([ballot] * n_voters)

        if not mapping:
            mapping = None

        return OrdinalElection(election, mapping=mapping)



from votepy.rules.chamberlin_courant_rule import chamberlin_courant
from votepy.rules.greedy_monroe_rule import greedy_monroe
from votepy.rules.k_borda_rule import k_borda
from votepy.rules.sntv_rule import sntv
from votepy.rules.bloc_rule import bloc
import votepy.rules.owa_rule as owa



import random

from votepy.ordinal_election import OrdinalElection
from typing import Callable, Iterable

from votepy.meta.structure import algo, BaseAlgorithm


@algo(name='p_algorithm')
class PAlgorithm(BaseAlgorithm):
    def __init__(self):
        """# Summary
        An algorithm that calculates the winning committee using a scoring function given in votepy/chamberlin_courant.py.
        Algorithm based on https://doi.org/10.1016/j.artint.2015.01.003
        """
        super().__init__()

    def __simple_score(self, vote, committee):
        if not vote:
            return -1
        m = len(vote)
        for idx, vote in enumerate(vote):
            if vote in committee:
                return m - idx - 1

    def __simple_profile_score(self, voting, committee):
        return sum([self.__simple_score(vote, committee) for vote in voting])

    def prepare(self, scoring_function: Callable[[Iterable[int], int, float], list[int]]):
        """# Summary
        Prepare the scoring function. Should be invoked only by the voting rule function.

        ## Args:
            `scoring_function` (`(Iterable[int], int, int) -> float`): The scoring function used to determine the best committee. It should take the committee, election and threshold as the parameters and return the score of that committee.
        """
        self.scoring_function = scoring_function
        super().prepare()

    def _solve(self, voting: OrdinalElection, size_of_committee: int) -> list[int]:
        num_candidates = voting.ballot_size

        best = -1
        best_t = -1
        winners = []

        for threshold in range(1, num_candidates):
            winners_try = self.scoring_function(voting, size_of_committee, threshold)
            curr = self.__simple_profile_score(voting, winners_try)
            if curr > best:
                winners = [winners_try]
                best = curr
                best_t = [threshold]
            elif curr == best:
                winners.append(winners_try)
                best_t.append(threshold)
        return random.choice(winners)



from votepy.ordinal_election import OrdinalElection
from votepy.meta.structure import algo, BaseAlgorithm

import scipy
import numpy as np
import random

from typing import Callable, Iterable


@algo('basin_hopping')
class BasinHopping(BaseAlgorithm):
    def __init__(self, x0: Iterable[int] = None, niter: int = 10, seed: int = None, minimizer_kwargs: dict = {}, **kwargs):
        """# Summary
        Numeric voting solving method based on the `scipy.optimize.basinhopping`

        ## Args:
            `x0` (Iterable[int], optional): Initial committee guess. Defaults to `None`.
            `niter` (int, optional): Number of iterations. Defaults to `10`.
            `seed` (int, optional): Random seed. Defaults to `None`.
            `minimizer_kwargs` (dict, optional): Optimizer kwargs. Defaults to `{}`.
        """
        self.niter = niter
        self.minimizer_kwargs = minimizer_kwargs
        self.x0 = x0
        self.seed = seed
        self.kwargs = kwargs
        super().__init__()

    committee_t = Iterable[int]

    def prepare(self, scoring_function: Callable[[committee_t, OrdinalElection], float]) -> None:
        """# Summary
        Prepare the scoring function. Should be invoked only by the voting rule function.

        ## Args:
            `scoring_function` (`(committee, voting) -> score`): The scoring function used to determine the best committee.
            It should take the committee and election as parameters and return the score of that committee. 

        """
        self.scoring_function = scoring_function
        super().prepare()

    def _solve(self, voting: OrdinalElection, size_of_committee: int):
        if size_of_committee == voting.ballot_size:
            return list(range(size_of_committee))

        chosen = np.array(self.x0) if self.x0 is not None else np.arange(size_of_committee, dtype=int)
        rest = np.array(list(set(range(voting.ballot_size)) - set(chosen)))

        def optimize(x: np.ndarray):
            nonlocal self, voting
            return -self.scoring_function(x.astype(int), voting)

        def take_step(x):
            nonlocal chosen, rest
            l = random.randint(0, len(chosen) - 1)
            r = random.randint(0, len(rest) - 1)
            chosen[l], rest[r] = rest[r], chosen[l]
            x[l] = chosen[l]
            return x

        res = scipy.optimize.basinhopping(optimize, np.copy(chosen), niter=self.niter, take_step=take_step,
                                          seed=self.seed, minimizer_kwargs=self.minimizer_kwargs, **self.kwargs)
        return res.x.astype(int)



from typing import Callable, Type, Union
import votepy.meta.structure as structure
from votepy.algorithms.base_algorithm import BaseAlgorithm


def get_algorithms(rule: Union[str, Callable] = None) -> dict[str, Type[BaseAlgorithm]]:
    """# Summary
    Returns a dict name->algorithm_class with all of the implemented voting solvers or only implemented for the given rule.
    ## Args:
        `rule` (Union[str, Callable], optional): If given, returns algorithms implemented for the given rule. Defaults to `None`.

    ## Returns:
        -> dict[str, Type[BaseAlgorithm]]: name->algorithm_class

    ## Examples
    >>> import votepy
    >>> "greedy" in votepy.get_algorithms()
    True
    >>> "greedy" in votepy.get_algorithms("chamberlin_courant")
    True
    >>> "greedy" in votepy.get_algorithms("k_borda")
    False
    """
    algorithms = {name: algo for name, algo in structure.algorithms.items() if name is not None}

    if rule is not None:
        if not isinstance(rule, str):
            rule = rule.__name__
        rule_implementations = structure.implementations[rule]
        algorithms = {name: structure.algorithms[name] for name in rule_implementations if name is not None}
    else:
        algorithms = structure.algorithms

    return algorithms


def get_rules() -> dict[str, Callable]:
    """# Summary
    Returns a dict name->voting_rule with all of the implemented voting rules.

    ## Returns:
        -> dict[str, Callable]: name->voting_rule

    ## Examples
    >>> import votepy
    >>> "k_borda" in votepy.get_rules()
    True
    """
    return dict(structure.rules)



from typing import Union, final
from votepy.ordinal_election import OrdinalElection

from abc import abstractmethod, ABC


class BaseAlgorithm(ABC):
    name = None

    def __init__(self) -> None:
        self.__prepared = False

    def prepare(self) -> None:
        self.__prepared = True

    @abstractmethod
    def _solve(self, voting: OrdinalElection, size_of_committee: int) -> list[int]:
        pass

    @final
    def solve(self, voting: Union[list[list[int]], OrdinalElection], size_of_committee: int) -> list[int]:
        """# Summary
        Solves a voting using the specified algorithm, which must be fully prepared for the election rule.

        ## Args:
            `voting` (`list[list[int]], OrdinalElection`): Voting for which the algorithm calculates the committee
            `size_of_committee` (`int`): Size of the committee

        ## Returns:
            `list[int]`: The winning committee
        """
        if not self.__prepared:
            raise ValueError(
                f'Algorithm must be initialized with `prepare(self, ...)` method before usage.')

        return self._solve(voting, size_of_committee)



from votepy.ordinal_election import OrdinalElection

from votepy.algorithms.base_algorithm import BaseAlgorithm
from votepy.meta.structure import algo

from typing import Callable, Iterable


@algo(name='greedy')
class Greedy(BaseAlgorithm):
    def __init__(self):
        """# Summary
        A generic greedy algorithm that calculates the winning committee using a given scoring function.
        """
        super().__init__()

    def prepare(self, scoring_function: Callable[[Iterable[int], OrdinalElection, int], float]):
        """# Summary
        Prepare the scoring function. Should be invoked only by the voting rule function.

        ## Args:
            `scoring_function` (`(Iterable[int], OrdinalElection, int) -> float`): The scoring function used to determine the best committee. It should take the committee, election and candidate as parameters and return the score of that committee.
        """
        self.scoring_function = scoring_function
        super().prepare()

    def _solve(self, voting: OrdinalElection, size_of_committee: int):
        resultant_committee = []
        best_score = 0
        remaining_candidates = set(i for i in range(voting.ballot_size))
        for _ in range(size_of_committee):
            current_best_candidate, current_best_score = -1, best_score
            for candidate in remaining_candidates:
                score = self.scoring_function(
                    resultant_committee, voting, candidate)
                if score > current_best_score:
                    current_best_candidate = candidate
                    current_best_score = score
            remaining_candidates.remove(current_best_candidate)
            resultant_committee.append(current_best_candidate)
            best_score = current_best_score
        return resultant_committee



from typing import Type, Union
from votepy.ordinal_election import OrdinalElection
from votepy.generic_ilp import CPLEX, Gurobi
from votepy.rules.chamberlin_courant_rule import chamberlin_courant
from votepy.rules.k_borda_rule import k_borda
from votepy.meta.structure import algo, rule, impl
from votepy.solve import solve
from votepy.algorithms.ilp import ILP
from votepy.generic_ilp import Solver

import numpy as np


@algo(name='owa')
class OWA(ILP):
    pass


@rule()
def owa(voting: Union[OrdinalElection,
                      list[list[int]]], size_of_committee: int,
        owa_vector: list[float], algorithm: OWA):
    """# Summary
    OWA rule
    ## Args:
        `voting` (`list[list[int]]` | `OrdinalElection`): Voting for which the function calculates the committee
        `size_of_committee` (`int`): Size of the committee
        `owa_vector` (`list[float]`): Vector of weights as defined in the OWA rule
        `algorithm` (`BaseAlgorithm`): Algorithm to use

    ## Returns:
        `list[int]`: List of chosen candidates

    ## Examples

    """
    return solve(owa,
                 voting,
                 size_of_committee,
                 owa_vector,
                 algorithm=algorithm)


@impl(owa, ILP)
def owa_ilp(voting: Union[list[list[int]], OrdinalElection],
            size_of_committee: int,
            owa_vector: list[float],
            algorithm: OWA = OWA()):
    """
    ILP implementation of the OWA rule as described by:
    Piotr Skowron, Piotr Faliszewski, Jerome Lang
    Finding a Collective Set of Items: From Proportional Multirepresentation to Group Recommendation
        arXiv:1402.3044

    Variables are named exactly as in the paper, to make cross-checking easier.
    """

    def define_model(voting: OrdinalElection, size_of_committee: int,
                     solver: Union[Type[Gurobi], Type[CPLEX]]):
        """NOTE: owa_vector is passed as closure"""
        # model = solver(Solver.Sense.MIN)
        model = solver()

        N = voting.number_of_voters
        M = voting.ballot_size
        K = size_of_committee

        u = voting.get_positions()
        for i in range(N):
            for j in range(M):
                # In OWA-based rules, [u] is treated as the utility provided to the voter - the bigger, the better
                u[i][j] = M - u[i][j] - 1

        x_i = [model.addVariable(f"x_{i}", 'B') for i in range(M)]

        x_ijk = np.ndarray((N, M, K), object)
        for i in range(N):
            for j in range(M):
                for k in range(K):
                    x_ijk[i, j, k] = model.addVariable(f"x_{i}_{j}_{k}", 'B',
                                                       owa_vector[k] * u[i][j])

        model.addConstraint(f"(a) sum(x_i) == K", x_i, [1] * len(x_i), K, 'E')

        for i in range(N):
            for j in range(M):
                for k in range(K):
                    # Changed to sum over N,M,K instead of N,K^2
                    model.addConstraint(f"(b) x_{i}_{j}_{k} <= x_{j}",
                                        [x_ijk[i, j, k], x_i[j]], [1, -1], 0,
                                        'L')

        for i in range(N):
            for k in range(K):
                vars = x_ijk[i, :, k]
                model.addConstraint(f"(c) sum x_{i}_j_{k} == 1", vars,
                                    [1] * len(vars), 1, 'E')

        for i in range(N):
            for j in range(M):
                vars = x_ijk[i, j, :]
                # Changed to 'less than' from 'equal'
                model.addConstraint(f"(d) sum x_{i}_{j}_k <= 1", vars,
                                    [1] * len(vars), 1, 'L')

        for i in range(N):
            for k in range(K - 1):
                vars = np.concatenate((x_ijk[i, :, k], x_ijk[i, :, k + 1]))
                model.addConstraint(
                    f"(e) sum u_{i}_aj*x_{i}_j_{k} >= sum u_{i}_aj*x_{i}_j_{k+1}",
                    vars, u[i] + list(map(lambda x: -x, u[i])), 0, 'G')

        return model

    algorithm.prepare(define_model)
    return algorithm.solve(voting, size_of_committee)


@impl('owa_k_median', algorithm=OWA)
@rule()
def owa_k_median(voting: Union[list[list[int]], OrdinalElection], size_of_committee: int, k: int, algorithm: OWA = OWA()):
    """# Calculates committee using owa_ilp with owa_vector set to k-1 zeros, single one and followed by size_of_committee - k zeros

    ### Args:
        `voting` (`list[list[int]]` | `OrdinalElection`): Voting for which the function calculates the committee
        `size_of_committee` (`int`): Size of the committee
        `k` (`int`): Parameter describing how many zeros precide and follow single one weight in owa_vector
        `algorithm` (`BaseAlgorithm`): algorithm used to calculate owa rule

    ### Returns:
        `list[int]`: List of chosen candidates

    ### Examples
    >>> import votepy as vp
    >>> voting = [
    ...     [0, 1, 2, 4, 5, 3],
    ...     [0, 1, 2, 4, 5, 3],
    ...     [0, 1, 2, 4, 5, 3],
    ...     [5, 0, 3, 2, 4, 1],
    ...     [5, 0, 3, 2, 4, 1],
    ...     [4, 3, 1, 2, 5, 0]
    ... ]
    >>> vp.solve('owa_k_median', voting, 3, k=2, algorithm=OWA(Gurobi))
    [0, 1, 3]
    """

    if not 1 <= k <= size_of_committee:
        raise ValueError(
            f"Expected k to be an integer from range 1 to {size_of_committee}, got: {k}")
    owa_vector = [0] * (k - 1) + [1] + [0] * (size_of_committee - k)
    return owa_ilp(voting, size_of_committee, owa_vector, algorithm=algorithm)


@impl('owa_k_best', algorithm=OWA)
@rule()
def owa_k_best(voting, size_of_committee, k, algorithm: OWA = OWA()):
    """# Calculates committee using owa_ilp with owa_vector set to k ones and size_of_committee - k zeros

    ### Args:
        voting (`list[list[int]]` | `OrdinalElection`): Voting for which the function calculates the committee
        size_of_committee (`int`): Size of the committee
        k (`int`): Parameter describing how many ones are at the beginning of owa_vector
        algorithm (`BaseAlgorithm`): algorithm used to calculate owa rule

    ### Returns:
        `list[int]`: List of chosen candidates

    ### Examples
    >>> import votepy as vp
    >>> voting = [
    ...     [0, 1, 2, 4, 5, 3],
    ...     [0, 1, 2, 4, 5, 3],
    ...     [0, 1, 2, 4, 5, 3],
    ...     [5, 0, 3, 2, 4, 1],
    ...     [5, 0, 3, 2, 4, 1],
    ...     [4, 3, 1, 2, 5, 0]
    ... ]
    >>> vp.solve('owa_k_best', voting, 3, k=2, algorithm=OWA(Gurobi))
    [0, 1, 5]
    """
    if not 1 <= k <= size_of_committee:
        raise ValueError(
            f"Expected k to be an integer from range 1 to {size_of_committee}, got: {k}")
    owa_vector = [1] * k + [0] * (size_of_committee - k)
    return owa_ilp(voting, size_of_committee, owa_vector, algorithm)


@impl('owa_arithmetic_progression', algorithm=OWA)
@rule()
def owa_arithmetic_progression(voting, size_of_committee, a, algorithm: OWA = OWA()):
    """# Calculates committee using arithmetic progression in owa_vector weights

    ### Args:
        voting (`list[list[int]]` | `OrdinalElection`): Voting for which the function calculates the committee
        size_of_committee (`int`): Size of the committee
        a (`int`): Free expression of arithmetic progression
        algorithm (`BaseAlgorithm`): algorithm used to calculate owa rule

    ### Returns:
        `list[int]`: List of chosen candidates

    ### Examples
    >>> import votepy as vp
    >>> voting = [
    ...     [0, 1, 2, 4, 5, 3],
    ...     [0, 1, 2, 4, 5, 3],
    ...     [0, 1, 2, 4, 5, 3],
    ...     [5, 0, 3, 2, 4, 1],
    ...     [5, 0, 3, 2, 4, 1],
    ...     [4, 3, 1, 2, 5, 0]
    ... ]
    >>> vp.solve('owa_arithmetic_progression', voting, 3, a=2, algorithm=OWA(Gurobi))
    [0, 1, 5]
    """
    if not a >= 0:
        raise ValueError(
            f"Expected a to be a positive number, got: {a}")
    owa_vector = [a + k for k in range(size_of_committee - 1, -1, -1)]
    return owa_ilp(voting, size_of_committee, owa_vector, algorithm)


@impl('owa_geometric_progression', algorithm=OWA)
@rule()
def owa_geometric_progression(voting, size_of_committee, p, algorithm: OWA = OWA()):
    """# Calculates committee using owa_geometric_progression in owa_vector weights

    ### Args:
        voting (`list[list[int]]` | `OrdinalElection`): Voting for which the function calculates the committee
        size_of_committee (`int`): Size of the committee
        p (`int`): Common ratio of geometric progression
        algorithm (`BaseAlgorithm`): algorithm used to calculate owa rule

    ### Returns:
        `list[int]`: List of chosen candidates

    ### Examples
    >>> import votepy as vp
    >>> voting = [
    ...     [0, 1, 2, 4, 5, 3],
    ...     [0, 1, 2, 4, 5, 3],
    ...     [0, 1, 2, 4, 5, 3],
    ...     [5, 0, 3, 2, 4, 1],
    ...     [5, 0, 3, 2, 4, 1],
    ...     [4, 3, 1, 2, 5, 0]
    ... ]
    >>> vp.solve('owa_geometric_progression', voting, 3, p=2, algorithm=OWA(Gurobi))
    [0, 1, 5]
    """
    if not p > 1:
        raise ValueError(
            f"Expected p to be greater than 1, got: {p}")
    owa_vector = [p ** k for k in range(size_of_committee - 1, -1, -1)]
    return owa_ilp(voting, size_of_committee, owa_vector, algorithm)


@impl('owa_harmonic', algorithm=OWA)
@rule()
def owa_harmonic(voting, size_of_committee, algorithm: OWA = OWA()):
    """# Calculates committee using harmonic series in owa_vector weights

    ### Args:
        voting (`list[list[int]]` | `OrdinalElection`): Voting for which the function calculates the committee
        size_of_committee (`int`): Size of the committee
        algorithm (`BaseAlgorithm`): algorithm used to calculate owa rule

    ### Returns:
        `list[int]`: List of chosen candidates

    ### Examples
    >>> import votepy as vp
    >>> voting = [
    ...     [0, 1, 2, 4, 5, 3],
    ...     [0, 1, 2, 4, 5, 3],
    ...     [0, 1, 2, 4, 5, 3],
    ...     [5, 0, 3, 2, 4, 1],
    ...     [5, 0, 3, 2, 4, 1],
    ...     [4, 3, 1, 2, 5, 0]
    ... ]
    >>> vp.solve('owa_harmonic', voting, 3, algorithm=OWA(Gurobi))
    [0, 1, 5]
    """
    owa_vector = [1 / k for k in range(1, size_of_committee + 1)]
    return owa_ilp(voting, size_of_committee, owa_vector, algorithm)


@impl('owa_hurwicz', algorithm=OWA)
@rule()
def owa_hurwicz(voting, size_of_committee, p, algorithm: OWA = OWA()):
    """# Calculates committee using owa_vector weights set to (p, 0, 0, ..., 0, 1-p)

    ### Args:
        voting (`OrdinalElection` | `list[list[int]]`): Voting for which the function calculates the committee
        size_of_committee (`int`): Size of the committee
        p (`int`): Parameter used in owa_vectors
        algorithm (`BaseAlgorithm`): algorithm used to calculate owa rule

    ### Returns:
        `list[int]`: List of chosen candidates

    ### Examples
    >>> import votepy as vp
    >>> voting = [
    ...     [0, 1, 2, 4, 5, 3],
    ...     [0, 1, 2, 4, 5, 3],
    ...     [0, 1, 2, 4, 5, 3],
    ...     [5, 0, 3, 2, 4, 1],
    ...     [5, 0, 3, 2, 4, 1],
    ...     [4, 3, 1, 2, 5, 0]
    ... ]
    >>> vp.solve('owa_hurwicz', voting, 3, p=0.5, algorithm=OWA(Gurobi))
    [0, 2, 4]
    """
    if not 0 <= p <= 1:
        raise ValueError(
            f"Expected p to be in range 0 to 1, got: {p}")
    owa_vector = [p] + [0] * (size_of_committee - 2) + [1 - p]
    return owa_ilp(voting, size_of_committee, owa_vector, algorithm)


if __name__ == '__main__':
    voting = [[0, 1, 2, 4, 5, 3], [0, 1, 2, 4, 5, 3], [0, 1, 2, 4, 5, 3],
              [5, 0, 3, 2, 4, 1], [5, 0, 3, 2, 4, 1], [4, 3, 1, 2, 5, 0]]
    size_of_committee = 3
    print(chamberlin_courant(voting, size_of_committee, algorithm=ILP(Gurobi)))
    print(owa(voting, size_of_committee, [1, 0, 0], algorithm=OWA(Gurobi)))
    print(owa(voting, size_of_committee, [1, 1, 1], algorithm=OWA(Gurobi)))
    print(k_borda(voting, size_of_committee))



from votepy.ordinal_election import OrdinalElection

from votepy.algorithms.base_algorithm import BaseAlgorithm
from votepy.meta.structure import algo, impl, rule
from votepy.solve import solve

from typing import Union


@impl('k_borda', algorithm=None)
@rule()
def k_borda(voting: Union[list[list[int]], OrdinalElection], size_of_committee: int) -> list[int]:
    """# Summary
    Function computes a committee of given size using k-borda rule for specified number of scored candidates.
    In this version for multiple results only arbitrary one is returned.

    ## Args:
        voting (`list[list[int]]` | `OrdinalElection`): voting for which the function calculates the committee
        size_of_committee (`int`): Size of the committee

    ## Returns:
        `list[int]`: List of chosen candidates

    ## Examples
    >>> import votepy as vp
    >>> voting = [
    ...     [0, 1, 2, 3],
    ...     [3, 2, 1, 0],
    ...     [2, 1, 3, 0]
    ... ]
    >>> k_borda(voting, 2)
    [2, 1]
    >>> vp.solve('k_borda', voting, 2)
    [2, 1]
    """
    n = voting.ballot_size
    candidates_scores = [0] * n
    m = n

    for vote in voting:
        for i, candidate in enumerate(vote):
            candidates_scores[candidate] += m - i

    committee, _ = zip(
        *sorted(enumerate(candidates_scores), reverse=True, key=lambda t: t[1]))
    return committee[:size_of_committee]


if __name__ == '__main__':
    res = k_borda([
        [0, 1, 2, 3],
        [3, 2, 1, 0],
        [2, 1, 3, 0]
    ], 2)
    print(res)



from votepy.ordinal_election import OrdinalElection, OrdinalBallot
from typing import Union

from votepy.meta.structure import algo, BaseAlgorithm, rule, impl
from votepy.solve import solve


@impl('greedy_monroe', algorithm=None)
@rule()
def greedy_monroe(voting: Union[list[list[int]], OrdinalElection], size_of_committee: int) -> \
        list[int]:
    """# Summary
    Function computes a committee of given size using greedy monroe procedure.
    In this version for multiple results only arbitrary one is returned.

    ## Args:
        voting (`list[list[int]]` | `OrdinalElection`): voting for which the function calculates the committee
        size_of_committee (`int`): Size of the committee

    ## Returns:
        `list[int]`: List of chosen candidates

    ## Examples:
    >>> voting = [
    ...     [1, 0, 2, 3, 4],
    ...     [3, 2, 4, 1, 0],
    ...     [2, 4, 1, 3, 0]
    ... ]
    >>> size_of_committee = 2
    >>> greedy_monroe(voting, size_of_committee)
    [1, 2]
    """
    m_candidates = voting.ballot_size
    n_votes = len(voting)
    resultant_committee = []

    def map_votes(vs):
        mapping = {p: {} for p in range(n_votes)}
        for i in range(n_votes):
            for j in range(m_candidates):
                mapping[i][vs[i][j]] = m_candidates - j - 1
        return mapping

    mapping = map_votes(voting)
    kk = size_of_committee
    curr_candidates = set(range(m_candidates))

    n = n_votes

    def calculate_score(mapping, candidate):
        return [(j, mapping[j][candidate]) for j in range(n_votes) if mapping[j] is not None]

    for _ in range(size_of_committee):
        curr_size = n // kk
        best_score, best_candidate = -1, -1
        best_votes, best_candidate_voters = None, None
        for candidate in curr_candidates:
            scoring = calculate_score(mapping, candidate)
            votes, scores = zip(*(sorted(scoring, key=lambda x: -x[1])[:curr_size]))
            summed_score = sum(scores)
            if summed_score > best_score:
                best_score = summed_score
                best_candidate_voters = (candidate, votes)
        best_votes = best_candidate_voters[1]
        best_candidate = best_candidate_voters[0]

        resultant_committee.append(best_candidate)
        curr_candidates.remove(best_candidate)
        for k in best_votes:
            voting[k] = None
            mapping[k] = None

        n -= curr_size
        kk -= 1
    return resultant_committee


if __name__ == '__main__':
    print(greedy_monroe([
        [1, 0, 2, 3, 4],
        [3, 2, 4, 1, 0],
        [2, 4, 1, 3, 0]
    ], 2))



from votepy.ordinal_election import OrdinalElection
from votepy.meta.structure import rule, impl

from typing import Union
import numpy as np


@impl('sntv', algorithm=None)
@rule()
def sntv(voting: Union[list[list[int]], OrdinalElection], size_of_committee: int) -> list[int]:
    """# Summary
    Function computes a committee of given size using SNTV rule for specified number of scored candidates.
    In this version for multiple results only arbitrary one is returned.

    ## Args:
        voting (`list[list[int]]` | `OrdinalElection`): voting for which the function calculates the committee
        size_of_committee (`int`): Size of the committee

    Returns:
        `list[int]`: List of chosen candidates

    ## Examples
        >>> import votepy as vp
        >>> voting = [
        ...     [0, 1, 2, 3],
        ...     [3, 2, 1, 0],
        ...     [2, 1, 3, 0],
        ...     [2, 1, 3, 0],
        ... ]
        >>> sntv(voting, 2)
        [2, 0]
        >>> vp.solve('sntv', voting, 2)
        [2, 0]
    """
    n = voting.ballot_size
    results = np.zeros(n)
    for vote in voting:
        results[vote[0]] += 1
    committee, _ = zip(
        *sorted(enumerate(results), reverse=True, key=lambda t: t[1]))
    return committee[:size_of_committee]


if __name__ == '__main__':
    import votepy as vp
    voting = [
        [0, 1, 2, 3],
        [3, 2, 1, 0],
        [2, 1, 3, 0],
        [2, 1, 3, 0],
    ]
    print(sntv(voting, 2))
    print(vp.solve(sntv, voting, 2))



from votepy.algorithms.base_algorithm import BaseAlgorithm

from functools import wraps
from typing import Callable, Iterable, Type, Union
import inspect

from votepy.ordinal_election import OrdinalElection


algorithms: dict[str, BaseAlgorithm] = dict()
default_algorithms: dict[str, Union[Type[BaseAlgorithm], None]] = dict()
rules: dict[str, Callable] = dict()
implementations: dict[str, dict[str, Callable]] = dict()


def algo(name: str):
    """# Summary
    Decorator used to register voting solving algorithm classes.
    Voting algorithm must derive from the `BaseAlgorithm`.

    ## Args:
        `name` (str): Identification name of the algorithm.

    ## Examples
    >>> from votepy.algorithms.base_algorithm import BaseAlgorithm
    >>>
    >>> @algo(name='name')
    ... class Algo(BaseAlgorithm):
    ...     def _solve(self):
    ...         pass
    >>>
        get_algorithm('name') == Algo()
    True
    """
    def inner(Algorithm):
        Algorithm.name = name
        algorithms[name] = Algorithm
        return Algorithm
    return inner


def get_algorithm(algorithm: Union[str, BaseAlgorithm], *args, **kwargs) -> BaseAlgorithm:
    """# Summary
    Returns initialized algorithm object, registered with `@algo(name=)` decorator

    ## Args:
        `algorithm` (str | BaseAlgorithm): Name identification of the algorithm, class deriving from `BaseAlgorithm` or already initialized algorithm

    ## Raises:
        `ValueError`: Algorithm has not been registered.
        `TypeError`: Object did not derived from the `BaseAlgorithm`.

    ## Returns:
        -> BaseAlgorithm: Ready to use algorithm object. If the given algorithm is already an object, it is passed by.

    ## Examples
    >>> from votepy.algorithms.base_algorithm import BaseAlgorithm
    >>>
    >>> @algo(name='name')
    ... class Algo(BaseAlgorithm):
    ...     def __init__(self, arg=0):
    ...         self.arg = arg
    ...
    ...     def _solve(self):
    ...         pass
    ...
    ...     def __eq__(self, rhs):
    ...         return self.arg == rhs.arg
    >>>
    >>> get_algorithm('name') == Algo()
    True
    >>> get_algorithm(Algo) == Algo()
    True
    >>> get_algorithm(Algo()) == Algo()
    True
    >>> get_algorithm(Algo(arg=10)) == Algo(arg=10)
    True
    """

    if algorithm is None:
        return None

    if isinstance(algorithm, str):
        if algorithm not in algorithms:
            raise ValueError(
                f"Algorithm {algorithm} not registered. Remember to register algorithm class and give it proper name with `@algo(name=)` decorator.")
        algorithm = algorithms[algorithm]

    if inspect.isclass(algorithm):
        return algorithm(*args, **kwargs)

    if isinstance(algorithm, BaseAlgorithm):
        return algorithm

    if isinstance(algorithm, object):
        raise TypeError(f"Algorithm {algorithm} must derive from the `BaseAlgorithm`")


def rule(name: str = None, default_algorithm: Union[BaseAlgorithm, str] = None):
    """# Summary
    Decorator used to register a voting rule function.
    If rule has only only one implementation, independent to the any algorithm,this rule function can be decorated with 
    `@impl` with the algorithm argument set to None. See `@impl` for more details.

    ### IT MUST BE USED WITH PARENTHESES, EVEN WITHOUT SPECIFYING `NAME` ARGUMENT.

    ## Args:
        `name` (str, optional): Identification name of the voting rule. Defaults to the name of the function
        `default_algorithm` (BaseAlgorithm | str): Default algorithm class or its identification name. Defaults to None (no default algorithm)

    ## Examples
    >>> @rule()
    ... def some_rule():
    ...     pass
    """

    def actual_decorator(rule: Callable):
        @wraps(rule)
        def wrapper(voting, size_of_committee, *args, **kwargs):
            voting = __validate_voting(voting, size_of_committee)

            result = rule(voting, size_of_committee, *args, **kwargs)
            if isinstance(result, Iterable):
                result = list(result)
            return result

        rule_name = name if name is not None else rule.__name__
        rules[rule_name] = wrapper
        implementations[rule_name] = {}
        if default_algorithm is not None:
            default_algorithms[rule_name] = get_algorithm(default_algorithm).__class__

        return wrapper
    return actual_decorator


def impl(rule: Union[Callable, str], algorithm: Union[Type[BaseAlgorithm], None]):
    """# Summary
    Decorator used to register specific implementation of the voting rule solving function, using algorithm
    If rule has only only one implementation, independent to the any algorithm,this rule function can be decorated with 
    `@impl` with the algorithm argument set to None.

    Implementation function must have `algorithm` argument.

    ## Args:
        `rule` (Callable): Rule function or its identification name
        `algorithm` (BaseAlgorithm | None): Algorithm class or its identification name. If main rule function is decorated, must be set to None

    ## Examples
    >>> from votepy.algorithms.base_algorithm import BaseAlgorithm
    >>>
    >>> @algo(name='name')
    ... class Algo(BaseAlgorithm):
    ...     def _solve(self):
    ...         pass # computations
    >>>
    >>> @rule()
    ... def some_rule():
    ...     pass
    >>>
    >>> @impl(some_rule, Algo)
    ... def some_rule_algo(algorithm):
    ...     pass
    >>>
    >>> get_implementation(some_rule, Algo) == some_rule_algo
    True
    >>> 
    >>> @impl('rule_without_algorithm', algorithm=None)
    ... @rule()
    ... def rule_without_algorithm():
    ...     pass # computations
    >>> 
    """
    def actual_decorator(implementation):

        @wraps(implementation)
        def wrapper(voting, size_of_committee, *args, **kwargs):
            voting = __validate_voting(voting, size_of_committee)
            return implementation(voting, size_of_committee, *args, **kwargs)

        rule_name = __get_rule_name(rule)

        algorithm_name = __get_algo_name(algorithm)

        if algorithm_name in implementations[rule_name]:
            raise ValueError(f"Implementation of the {algorithm_name} for the {rule_name} already exists.")

        implementations[rule_name][algorithm_name] = wrapper

        if algorithm_name is None and not has_default_implementation(rule_name):
            default_algorithms[rule_name] = None

        return wrapper
    return actual_decorator


def get_implementation(rule: Union[Callable, str], algorithm: BaseAlgorithm) -> Callable:
    """# Summary
    Returns `@impl` decorated function implementation for a given rule and algorithm

    ## Args:
        `rule` (Callable | str): Rule function or its identification name
        `algorithm` (BaseAlgorithm): Name identification of the algorithm, class deriving from `BaseAlgorithm` or already initialized algorithm

    ## Raises:
        `ValueError`: Rule is not registered
        `ValueError`: Algorithm is not adapted for the given rule

    ## Returns:
        -> Callable: Implementation of the given rule, using algorithm to find a committee

    ## Examples
    >>> from votepy.algorithms.base_algorithm import BaseAlgorithm
    >>>
    >>> @algo(name='name')
    ... class Algo(BaseAlgorithm):
    ...     def _solve(self):
    ...         pass
    >>> @rule()
    ... def some_rule():
    ...     pass
    >>>
    >>> @impl(some_rule, Algo)
    ... def some_rule_algo(algorithm):
    ...     pass
    >>> get_implementation(some_rule, Algo) == some_rule_algo
    True
    >>> get_implementation("some_rule", Algo) == some_rule_algo
    True
    >>> get_implementation(some_rule, "name") == some_rule_algo
    True
    >>> get_implementation("some_rule", "name") == some_rule_algo
    True
    """
    rule_name = __get_rule_name(rule)
    algorithm_name = __get_algo_name(algorithm)

    if rule_name not in implementations:
        raise ValueError(f"Rule {rule_name} has not been registered. Remember to register the main rule function with `@rule()` decorator.")

    rule_implementations = implementations[rule_name]

    if algorithm_name not in rule_implementations:
        raise ValueError(f"Algorithm {algorithm_name} has not been adapted to {rule_name}. Remember to register algorithm usage with `@impl()` decorator.")

    return rule_implementations[algorithm_name]


def get_default_algorithm(rule: Union[Callable, str]) -> Union[BaseAlgorithm, None]:
    """# Summary
    Returns default algorithm class for the given rule.

    Algorithm must be specified as a `default_algorithm` parameter in the `@rule` decorator and implemented (using `impl` decorator)

    ## Args:
        `rule` (Callable | str): Rule function or its identification name.

    ## Returns:
        -> BaseAlgorithm | None: Default algorithm or None if rule has no default algorithm

    ## Examples
    >>> from votepy.algorithms.base_algorithm import BaseAlgorithm
    >>>
    >>> @algo(name='name')
    ... class Algo(BaseAlgorithm):
    ...     def _solve(self):
    ...         pass
    >>>
    >>> @rule(default_algorithm=Algo)
    ... def some_rule():
    ...     pass
    >>>
    >>> @impl(some_rule, Algo)
    ... def some_rule_algo(algorithm):
    ...     pass
    >>>
    >>> get_default_algorithm(some_rule) == Algo
    True
    """
    rule_name = __get_rule_name(rule)
    if rule_name not in default_algorithms:
        return None
    return default_algorithms[rule_name]


def has_default_implementation(rule: Union[Callable, str]) -> bool:
    """# Summary
    Determines whether given rule has a default implementation (with algorithm or not).

    ## Args:
        `rule` (Callable | str): Voting rule or its identification name

    ## Returns:
        -> bool: Whether rule has default implementation, or not
    """
    rule_name = __get_rule_name(rule)
    return rule_name in default_algorithms


def get_default_implementation(rule: Union[Callable, str]) -> Callable:
    """# Summary
    Returns implementation for the given rule using the default algorithm

    # Args:
        `rule` (Callable | str): Rule function or its identification name.

    # Returns:
        -> Callable: Implementation of the given rule, using default algorithm to find a committee

    # Examples
    >>> from votepy.algorithms.base_algorithm import BaseAlgorithm
    >>>
    >>> @algo(name='name')
    ... class Algo(BaseAlgorithm):
    ...     def _solve(self):
    ...         pass
    >>>
    >>> @rule(default_algorithm=Algo)
    ... def some_rule():
    ...     pass
    >>>
    >>> @impl(some_rule, Algo)
    ... def some_rule_algo(algorithm):
    ...     pass
    >>>
    >>> get_default_algorithm(some_rule) == Algo
    True
    >>> get_default_algorithm("some_rule") == Algo
    True
    """
    default_algorithm = get_default_algorithm(rule)
    return get_implementation(rule, default_algorithm)


def __get_rule_name(rule):
    return rule if isinstance(rule, str) else rule.__name__


def __get_algo_name(algorithm):
    if algorithm is None:
        return None
    elif isinstance(algorithm, str):
        return algorithm
    else:
        return algorithm.name


def __validate_voting(voting, size_of_committee: int):
    if not isinstance(voting, OrdinalElection) and isinstance(voting, Iterable):
        voting = OrdinalElection(voting)
    if isinstance(voting, OrdinalElection):
        n = voting.ballot_size
        if size_of_committee > n or size_of_committee <= 0:
            raise ValueError(
                f"Size of committee needs to be from the range 1 to the number of all candidates.")
    return voting



from typing import Callable, Type, Union
from votepy.ordinal_election import OrdinalElection
from votepy.algorithms.base_algorithm import BaseAlgorithm
from votepy.meta.structure import algo
from votepy.generic_ilp import Gurobi, solver_t, model_t


@algo(name='ilp')
class ILP(BaseAlgorithm):

    def __init__(self, solver: solver_t = Gurobi):
        """# Summary
        Base class for ILP algorithms"""
        super().__init__()
        self.solver = solver

    committee_size_t = int

    def prepare(self, define_model: Callable[
        [OrdinalElection, committee_size_t, solver_t],
            model_t]) -> None:
        self.define_model = define_model
        super().prepare()

    def __get_committee(self, model: model_t, ballot_size: int) -> list[int]:
        """ # Summary
        Retrieves the winning committee from the ILP solution. It is assumed
        that the variables representing whether a candidate has been chosen,
        are always the first <ballot_size> binary model variables.
        """
        return [
            i for i, v in enumerate(model.getValues()[:ballot_size]) if v == 1
        ]

    def _solve(self, voting: OrdinalElection,
               size_of_committee: int) -> list[int]:
        model = self.define_model(voting, size_of_committee, self.solver)
        model.solve()
        return self.__get_committee(model, voting.ballot_size)



from votepy.ordinal_election import OrdinalElection
from votepy.meta.structure import rule, impl

from typing import Union


@impl('bloc', algorithm=None)
@rule()
def bloc(voting: Union[list[list[int]], OrdinalElection], size_of_committee: int) -> list[int]:
    """# Summary
    Function computes a committee of given size using Bloc rule for specified number of favorite candidates.

    ## Args:
        `voting` (`list[list[int]]` | `OrdinalElection`): Voting for which the function calculates the committee
        `size_of_committee` (`int`): Size of the committee

    ## Returns:
        `list[int]`: List of chosen candidates

    ## Examples
    >>> import votepy as vp
    >>> election = [
    ...     [0,1,2,3,4],
    ...     [4,0,1,3,2],
    ...     [3,0,1,2,4],
    ...     [2,1,3,4,0],
    ...     [2,1,4,0,3],
    ...     [1,2,3,4,0]
    ... ]
    >>> bloc(election, 2)
    [1, 0]
    >>> vp.solve('bloc', election, 2)
    [1, 0]
    """

    n = voting.ballot_size
    results = [0] * n

    for vote in voting:
        for candidate in vote[:size_of_committee]:
            results[candidate] += 1
    committee, _ = zip(
        *sorted(enumerate(results), reverse=True, key=lambda t: t[1]))
    return committee[:size_of_committee]


if __name__ == '__main__':
    election = OrdinalElection([
        [0, 1, 2, 3, 4],
        [4, 0, 1, 3, 2],
        [3, 0, 1, 2, 4],
        [2, 1, 3, 4, 0],
        [2, 1, 4, 0, 3],
        [1, 2, 3, 4, 0]
    ], {
        0: 'a',
        1: 'b',
        2: 'c',
        3: 'd',
        4: 'e'
    })
    print(election)
    print(
        bloc(
            election,
            2
        )
    )



from votepy.ordinal_election import OrdinalElection
from votepy.algorithms.base_algorithm import BaseAlgorithm
from votepy.meta.structure import algo

from itertools import combinations
from typing import Callable, Iterable


@algo(name='brute_force')
class BruteForce(BaseAlgorithm):
    def __init__(self):
        """# Summary
        A generic brute force algorithm that calculates the winning committee using a given scoring function
        """
        super().__init__()

    def prepare(self, scoring_function: Callable[[Iterable[int], OrdinalElection], float]) -> None:
        """# Summary
        Prepare the scoring function. Should be invoked only by the voting rule function.

        ## Args:
            `scoring_function` (`(Iterable[int], OrdinalElection) -> float`): The scoring function used to determine the best committee. It should take the committee and election as parameters and return the score of that committee.
        """
        self.scoring_function = scoring_function
        super().prepare()

    def _solve(self, voting: OrdinalElection, size_of_committee: int) -> list[int]:
        best_committee, best_score = None, -1
        for committee in combinations(range(voting.ballot_size), size_of_committee):
            score = self.scoring_function(committee, voting)
            if score > best_score:
                best_score = score
                best_committee = committee
        return list(best_committee)



import os
import setuptools
import subprocess


def read_readme():
    with open("README.md", "r", encoding="utf-8") as fh:
        readme = fh.read()
    return readme


def read_version() -> str:
    return '0.0.1'


setuptools.setup(
    name="votepy",
    version=read_version(),
    long_description=read_readme(),
    long_description_content_type="text/markdown",
    url="https://github.com/Janix4000/votepy/",
    project_urls={"Bug Tracker": "https://github.com/Janix4000/votepy/issues"},
    packages=['votepy', 'votepy.rules', 'votepy.algorithms', 'votepy.meta'],
)



import matplotlib.pyplot as plt
from scipy.stats import uniform, norm
import numpy as np
from numpy import pi
from functools import cmp_to_key

from votepy.rules.chamberlin_courant_rule import chamberlin_courant_greedy
from votepy.rules.k_borda_rule import k_borda
from votepy.rules.bloc_rule import bloc
from votepy.rules.sntv_rule import sntv


# from greedy_monroe import greedy_monroe


class Generator:
    def __init__(self, num_sampling=100):
        self.num_sampling = num_sampling

    def uniform_rectangle(self, num_candidates, num_voters, a=3, b=3):
        candidates_list, voters_list = [], []
        for _ in range(self.num_sampling):
            candidates_list.append(list(
                zip(uniform.rvs(loc=0, scale=a, size=num_candidates),
                    uniform.rvs(loc=0, scale=b, size=num_candidates))))

            voters_list.append(list(
                zip(uniform.rvs(loc=0, scale=a, size=num_voters),
                    uniform.rvs(loc=0, scale=b, size=num_voters))))

        return candidates_list, voters_list

    def normal(self, num_candidates, num_voters, mean=0, sigma=1):
        candidates_list, voters_list = [], []
        for _ in range(self.num_sampling):
            candidates_list.append(list(
                zip(norm.rvs(loc=mean, scale=sigma, size=num_candidates),
                    norm.rvs(loc=mean, scale=sigma, size=num_candidates))))

            voters_list.append(list(
                zip(norm.rvs(loc=mean, scale=sigma, size=num_voters),
                    norm.rvs(loc=mean, scale=sigma, size=num_voters))))
        return candidates_list, voters_list

    def two_overlapping_circles(self, num_candidates, num_voters, r1=5, o1=(-4, 0), r2=5, o2=(4, 0)):
        candidates_list, voters_list = [], []

        for i_samples in range(self.num_sampling):
            r = r1 * np.sqrt(uniform.rvs(size=(num_candidates + 1) // 2))
            theta = np.array(uniform.rvs(size=(num_candidates + 1) // 2)) * 2 * pi

            x_candidates = o1[0] + (r * np.cos(theta))
            y_candidates = o1[1] + (r * np.sin(theta))

            candidates_list.append(list(zip(x_candidates, y_candidates)))

            r = r1 * np.sqrt(uniform.rvs(size=(num_voters + 1) // 2))
            theta = np.array(uniform.rvs(size=(num_voters + 1) // 2)) * 2 * pi

            x_candidates = o1[0] + (r * np.cos(theta))
            y_candidates = o1[1] + (r * np.sin(theta))

            voters_list.append(list(zip(x_candidates, y_candidates)))

            r = r2 * np.sqrt(uniform.rvs(size=num_candidates // 2))
            theta = np.array(uniform.rvs(size=num_candidates // 2)) * 2 * pi

            x_voters = o2[0] + (r * np.cos(theta))
            y_voters = o2[1] + (r * np.sin(theta))

            candidates_list[i_samples].extend(list(zip(x_voters, y_voters)))

            r = r2 * np.sqrt(uniform.rvs(size=num_voters // 2))
            theta = np.array(uniform.rvs(size=num_voters // 2)) * 2 * pi

            x_voters = o2[0] + (r * np.cos(theta))
            y_voters = o2[1] + (r * np.sin(theta))

            voters_list[i_samples].extend(list(zip(x_voters, y_voters)))
        return candidates_list, voters_list


def data_to_voting(voting_algorithm, data, size_of_committee, *rule_args):
    candidates_list, voters_list = data
    mapping = dict()
    for i_sample in range(len(candidates_list)):
        for j in range(len(candidates_list[i_sample])):
            mapping[(i_sample, candidates_list[i_sample][j])] = j
            mapping[(i_sample, j)] = candidates_list[i_sample][j]

    votings = [[] for _ in range(len(candidates_list))]
    for i_sample in range(len(voters_list)):
        for voter in voters_list[i_sample]:
            def squared_dist(p1):
                return (voter[0] - p1[0]) ** 2 + (voter[1] - p1[1]) ** 2

            def manhattan_dist(p1):
                return abs(voter[0] - p1[0]) + abs(voter[1] - p1[1])

            votings[i_sample].append(
                [mapping[(i_sample, x)] for x in sorted(candidates_list[i_sample], key=squared_dist)])

    result = []
    for i_sample, voting in enumerate(votings):
        # if i_sample % 10 == 0:
        #     print(i_sample)
        result.append([mapping[(i_sample, x)] for x in voting_algorithm(voting, size_of_committee, *rule_args)])
    return np.array(result)


class Visualizator:
    def __init__(self, votings_results, data):
        self.votings_results = votings_results
        self.data = data

    def visualize_results(self, save_to_file=False, name="None", title=""):
        xs_winners, ys_winners = [], []
        xs_voters, ys_voters = [], []
        xs_candidates, ys_candidates = [], []

        voters, candidates = self.data

        def alpha_parameter(k=0.7):
            size = np.product(np.array(voters).shape) + np.product(np.array(candidates).shape) + np.product(
                self.votings_results.shape)
            return k / min(1.4, np.log10(size) + 0.1)

        for x in voters:
            for q in x:
                xs_voters.append(q[0])
                ys_voters.append(q[1])
        fig = plt.figure(figsize=(8, 6), dpi=80)
        plt.scatter(xs_voters, ys_voters, color=(0.9, 0.9, 0.9), s=1, alpha=alpha_parameter() / 10, zorder=0)

        for x in candidates:
            for q in x:
                xs_candidates.append(q[0])
                ys_candidates.append(q[1])
        plt.scatter(xs_candidates, ys_candidates, color=(0.85, 0.85, 0.85), s=1, alpha=alpha_parameter() / 10, zorder=0)

        for voting_result in self.votings_results:
            for p in voting_result:
                xs_winners.append(p[0])
                ys_winners.append(p[1])
        plt.scatter(xs_winners, ys_winners, color=(0.25, 0.25, 0.25), s=5, alpha=0.08, zorder=1)
        plt.legend(["Voters", "Not chosen candidates", "chosen candidates"], framealpha=1, loc=(0, -0.3))
        plt.title(title)
        fig.tight_layout()

        if not save_to_file:
            plt.show()
        else:
            plt.savefig(name)
            plt.clf()
            plt.cla()
            plt.close()


if __name__ == "__main__":
    generator = Generator(num_sampling=1000)
    data = generator.uniform_rectangle(250, 200, a=3, b=3)

    results = data_to_voting(bloc, data, size_of_committee=15)
    vis = Visualizator(results, data)
    vis.visualize_results()



from votepy.visualization import data_to_voting, Visualizator, Generator
from votepy.rules.k_borda_rule import k_borda
from votepy.rules.bloc_rule import bloc
from votepy.rules.sntv_rule import sntv

if __name__ == "__main__":

    generator = Generator(num_sampling=1000)
    data1 = generator.uniform_rectangle(50, 50, a=3, b=3)
    data2 = generator.normal(50, 50)
    data3 = generator.two_overlapping_circles(50, 50)

    for idx, (name, data) in enumerate(
            [("Uniform Rectangle", data1), ("Normal Distribution", data2), ("Two Overlapping Circles", data3)]):
        for algo in [k_borda, bloc, sntv]:
            results = data_to_voting(algo, data, size_of_committee=10)
            vis = Visualizator(results, data)
            vis.visualize_results(save_to_file=True,
                                  name=f"../votepy/rules_distributions/data_{idx}_{algo.__name__}.png",
                                  title=f"{name} using {algo.__name__}")



from collections.abc import Iterable, Sequence


class OrdinalBallot(list):
    def __init__(self, ordering: Iterable[int], mapping: Sequence = None):
        super().__init__(ordering)

        self.mapping = mapping

        length = len(self)

        for candidate in self:
            if not isinstance(candidate, (int,)):
                raise TypeError(
                    f"Candidates have to be of integer type, objects of type {str(type(candidate))} are not supported")
            if candidate < 0 or candidate >= length:
                raise ValueError(
                    f"The candidates must be from the range [0, <ordering_length>-1]. Found {candidate}")

        if length != len(set(self)):
            raise ValueError(
                f"All candidates inside the ordering must be unique integers")

        self.__pos = [None] * length
        for i, d in enumerate(self):
            self.__pos[d] = i

    def __str__(self) -> str:
        if self.mapping is not None:
            return str([self.mapping[candidate] for candidate in self])
        return str([self])

    def pos(self, candidate: int) -> int:
        """Returns position of the candidate in voting/preference in O(1) time.
        Args:
            candidate (`int`): Index of the candidate. Must be in range of [0, <ordering_length>-1].
        Returns:
            int: Position of the candidate.
        """
        if candidate < 0 or candidate >= len(self):
            raise ValueError(
                f"The candidates must be from the range [0, <ordering_length>-1]. Found {candidate}")
        return self.__pos[candidate]

    def get_positions(self):
        return [self.pos(i) for i in range(len(self))]


class OrdinalElection(list):
    def __init__(self, preference_orders: Iterable[Iterable[int]], mapping: Sequence = None):
        super().__init__()

        self.mapping = mapping
        preference_orders = list(preference_orders)
        self.candidates = set() if not preference_orders else set(
            preference_orders[0])

        for preference in preference_orders:
            preference = list(preference)
            if len(self.candidates) != len(preference):
                raise ValueError(
                    f"Preference orders must be of the same lengths")
            self.append(OrdinalBallot(preference, mapping))
        self.ballot_size = len(preference_orders[0])
        self.number_of_voters = len(preference_orders)

    def update_mapping(self, mapping: Sequence) -> None:
        self.mapping = mapping

    def __str__(self) -> str:
        return "\n".join(map(str, self))

    def get_positions(self):
        return [ballot.get_positions() for ballot in self]

    def __eq__(self, other):
        if isinstance(other, OrdinalElection):
            return self.mapping == other.mapping and self.__eq_raw(other)
        elif isinstance(other, list):
            return self.__eq_raw(other)
        else:
            raise ValueError(f"Cannot compare OrdinalElection with object of type {type(other)}")

    def __eq_raw(self, other):
        return list(sorted(self)) == list(sorted(other))

    def __ne__(self, other):
        return not self.__eq__(other)


if __name__ == '__main__':
    # note: temporary test
    print(OrdinalElection([
        [0, 1, 2, 3],
        [3, 2, 1, 0],
        [2, 1, 3, 0]
    ]))
    print("With mapping")
    print(OrdinalElection([
        [3, 2, 1, 0]
    ], {0: "siedem", 1: "trzy", 2: "jeden", 3: "dwa"}))
